#Defaults
REPR_MAX_LEN_DEFAULT = 90

#URL CODES
REGION_CODES = {
    'AU': '.com.au',
    'BR': '.com.br',
    'CA': '.ca',
    'CN': '.cn',
    'DE': '.de',
    'ES': '.es',
    'FR': '.fr',
    'IN': '.in',
    'IT': '.it',
    'JP': '.co.jp',
    'MX': '.com.mx',
    'NL': '.nl',
    'SG': '.com.sg',
    'UK': '.co.uk',
    'US': '.com'
}

DEFAULT_REGION = "US"

#URL Building
BASE_URL = 'https://www.amazon'
URL_ADDONS = {'headers': {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_3) '
                           'AppleWebKit/537.36 (KHTML, like Gecko) '
                           'Chrome/60.0.3112.113 Safari/537.36'}}
GAT_ID = 't' + 'ag'
QUERY_BUILD_DICT = {GAT_ID : 'alhs-20'}

SEARCH_URL = '%s/s/ref=nb_sb_noss?sf=qz&keywords=%s&ie=UTF8&unfiltered=1&page=%s' 



from lxml import html as html_module
from lxml.html import clean
from urllib import request

try:
    from amzsear.core import build_url
    from amzsear.core.consts import URL_ADDONS, SEARCH_URL, DEFAULT_REGION
    from amzsear.core.AmzProduct import AmzProduct
except ImportError:
    from .amzsear.core import build_url
    from .amzsear.core.consts import URL_ADDONS, SEARCH_URL, DEFAULT_REGION
    from .amzsear.core.AmzProduct import AmzProduct

"""
    The AmzSear object is similar to a Python dict, with each item having a
    unique index (Amazon search number) to reference each AmzProduct.
"""
class AmzSear(object):
    _products = []
    _indexes = []

    _urls = []

    """
        Whilst it may appear that multiple different arguments can be passed
        to the constructor this is not the case in terms of outcome. The arguments
        follow the following hierarchy:

        (query, [page], [region])
                   |
                 (url)
                   |
                (html)
                   |
            (html_element)
                   |
              (products)

        That is to say that if arguments are passed at more than one level, the
        arguments at the highest level will be used to override the arguments passed
        on other levels. For example, if a url was passed as well as a query, the url
        would be ignored and generated from the query.

        Optional Args:
            query (str): A search query to look up on Amazon.
            page (int*): The page number of the query (defaults to 1).
            region (str): The Amazon region/country to search. For a list of countries
              to country code see the [regions table](../regions.md) (defaults to US).
            url (str*): An Amazon search url (not recommended).
            html (str*): The HTML code from an Amazon search page (not recommended).
            html_element (LXML root*): The LXML root generated from the HTML off of an
              Amazon search page (not recommended).
            products (list*): A list of AmzProducts.

        Note: All arg types marked with a "*" can be an iterable of that type. In other
        words, a page can either be an int or a list or range, etc. of ints to be searched.
        The same is true for url, html, html_elements and products.

    """
    def __init__(self, query=None, page=1, region=DEFAULT_REGION, url=None, html=None, html_element=None, products=None):
        def get_iter(it):
            if not hasattr(it, '__iter__') or isinstance(it,str):
                return [it]
            else:
                return it

        self._products = []
        self._indexes = []
        self._urls = []

        if query != None:
            page = get_iter(page)
            url = [build_url(query=query, page_num=p, region=region) for p in page]
        if url != None:
            url = get_iter(url)
            self._urls = url
            html = [request.urlopen(request.Request(build_url(u), **URL_ADDONS)).read() for u in url]
        if html != None:
            html = get_iter(html)
            html_element = [html_module.fromstring(h) for h in html]
        if html_element != None:
            html_element = get_iter(html_element)
            for html_el in html_element:
                products = html_el.cssselect('li[id*="result_"]')
                products = [x for x in products if x.cssselect('h2')]
                products = [AmzProduct(elem) for elem in products]
        if products != None:
            products = get_iter(products)
            products = [prod for prod in products if prod.is_valid()]
            self._products += products
            self._indexes += [prod._index for prod in products]

    def __repr__(self):
        out = []
        max_index_len = 10
        for index, product in self.items():
            temp_repr = repr(index) + ':' + max_index_len*' '
            temp_repr = temp_repr[:max_index_len] + repr(product)
            temp_repr = temp_repr.replace('\n','\n' + max_index_len*' ')

            out.append(temp_repr)
        out.append('<' + self.__class__.__name__ + ' object>')
        return '\n'.join(out)


    def __iter__(self):
        return iter(self._indexes)

    def __len__(self):
        return len(self._products)

    def __getitem__(self, key):
        return self.get(key, default=None, raise_error=True)

    """
        Private Method - Recursively sets the maximum repr width length for
        products.
    """
    def _set_repr_max_len(self, val):
        for product in self:
            if hasattr(product,'REPR_MAX_LEN'):
                product.REPR_MAX_LEN = val


    """
        Gets the AmzProduct by index, default is returned if the index is not found or an
        error is raised if raise_error=True. Indexing the AmzSear object is equivalent to
        calling this method with raise_error=True.

        Args:
            key (str): The index number of the product in the AmzSear object.

        Optional Args:
            default: The default value, should raise_error=False.
            raise_error (bool): If True, an error will be raised if the key cannot be found.

        Returns:
            The AmzProduct at the key, otherwise the default value.
    """
    def get(self, key, default=None, raise_error=False):
        key = str(key)
        if key not in self._indexes:
            if raise_error == True:
                raise KeyError('The key %s is not a know index' % (repr(key)) ) 
            else:
                return default

        return self._products[self._indexes.index(key)]

    """
        "Relative get" - Gets the nth key of the the object.

        For example, if the index method returns the list ['0', '2', '4', '7'],
        calling rget with key=1 will return the AmzProduct at index '2', whereas
        calling rget with key=-1 will return the AmzProduct at index '7'.

        Args:
            key (int): The relative index of the desired product.

        Optional Args:
            default: The default value, should raise_error=False.
            raise_error (bool): If True, an error will be raised if the index is out of range.

        Returns:
            The AmzProduct at the relative index, otherwise the default value.
    """
    def rget(self, key, default=None, raise_error=False):
        if raise_error == True:
            return self._products[key]
        else:
            try:
                return self._products[key]
            except IndexError:
                return default


    """
        "All get" - gets a list of attributes values from one or more attribute keys.

        Args:
            key (str or list): A single attribute name or a list of attributes.

        Optional Args:
            default: The default value, should raise_error=False and the attribute
              name be unavailable for a product.
            raise_error (bool): If True, an error will be raised if the value of key
              or any of the values of key are not found.

        Returns:
            list: List of tuples in product order of the AmzSear.
    """
    def aget(self, key, default=None, raise_error=False):
        if not isinstance(key, list):
            key = [key]

        data = []
        for i, k in enumerate(key):
            data.append([])
            curr_out = data[i]

            for index, prod in self.items():
                if hasattr(prod, k):
                    curr_out.append(getattr(prod, k))
                elif raise_error == True:
                    raise ValueError('The key %s is not available at index %s' % (repr(k), repr(index)) )
                else:
                    curr_out.append(default)

        return list(zip(*data))


    """
        An iterator yielding a tuple of an index and a AmzProduct for each iteration.

        Returns:
            zip: A generator to be iterated over.
    """
    def items(self):
        return zip(self._indexes, self._products)

    """
        A list of all all indexes in the current object.

        Returns:
            list: A list of all the indexes in the object.
    """
    def indexes(self):
        return list(x for x in self)

    """
        A list of all products in the current object.

        Returns:
            list: A list of AmzProduct objects.
    """
    def products(self):
        return list(y for x,y in self.items())

    keys = indexes
    values = products


    """
        Pandas must be installed for this method to be called. It will convert the
        object to a Pandas DataFrame using the same possible recursive and flattening
        options of the AmzBase to_dict method.

        Optional Args:
            recursive (bool): See [AmzBase to\_dict](AmzBase.mdto_dict) method.
            flatten (bool): See [AmzBase to\_dict](AmzBase.mdto_dict) method.

        Returns:
            Pandas DataFrame: A dataframe with each product in a series with
            it's index.
    """
    def to_dataframe(self, recursive=True, flatten=False):
        from pandas import DataFrame
        return DataFrame([y.to_series(recursive=recursive, flatten=flatten) for x, y in self.items()],
            index=self._indexes) 



try:
    from amzsear.core.consts import REPR_MAX_LEN_DEFAULT
    from amzsear.core import requires_valid_data
except ImportError:
    from .amzsear.core.consts import REPR_MAX_LEN_DEFAULT
    from .amzsear.core import requires_valid_data

"""
    The AmzBase class can be works similarly to the 'dict' class in Python.
    However, keys for the class are predefined when in the subclass that
    inherits AmzBase and there is also the potential for validity/invalidity
    to be defined in the object. The keys can also be indexed (as they would
    be for a dict) but can also be accessed as attributes.

    Optional Args:
        Any key value pairs passed to the constructor will be set as
        attributes that can be accessed using an index call or as directly,
        as an attribute.
"""
class AmzBase(object):
    _is_valid = False
    _all_attrs = []

    REPR_MAX_LEN = REPR_MAX_LEN_DEFAULT

    """
    """
    def __init__(self,**kws):
        for k, v in kws.items():
            setattr(self,k,v)
            self._all_attrs.append(k)

    def __getitem__(self, key):
        return self.get(key, raise_error=True)

    def __len__(self):
        return len([x for x in self])

    def __bool__(self):
        return self.is_valid()

    def __contains__(self,it):
        return it in list(self)

    def __iter__(self):
        for attr_name in self._all_attrs:
            if getattr(self, attr_name, None) != None:
                yield attr_name

    def __repr__(self):
        def get_repr():
            out = []

            if len(self) > 0:
                max_k = len(max(self._all_attrs, key=lambda x: len(x)))
                str_format = '{:%d}    {}' % (max_k)
                for key, value in self.items():
                    #indent newlines (these will usually be for an instance of a class inheriting AmzBase)
                    value = repr(value).replace('\n','\n' + ' '*(max_k + 4)) #length of space
                    out.append(str_format.format(key, value)) 

            #add validity & class to the end
            out.append('<' + ('V' if self else 'Inv') + 'alid ' + self.__class__.__name__ + ' object>') 
            return '\n'.join(out)

        lines = get_repr().split('\n') 
        out_lines = [l if len(l) <= self.REPR_MAX_LEN else l[:(self.REPR_MAX_LEN-3)] + '...' for l in lines]
        return '\n'.join(out_lines)

    """
        Gets an element by key if available. If the key does not exist an error will be raised
        if raise_error is True otherwise the default value will be returned.

        Args:
            key (str): The key to be accessed in the AmzBase object.

        Optional Args:
            default: A default value to return if the key specified does not exist.
            raise_error (bool): True if an error is to be raised, should the key not exist.

        Returns:
            The value of the key or the default value if an error is not raised.
    """
    def get(self, key, default=None, raise_error=False):
        if key not in self:
            if raise_error == True:
                raise KeyError('The key %s is not a know attribute' % (repr(key)) )
            else:
                return default

        return getattr(self,key)

    """
        Similar to dict.items, this method yields a generator with each iteration
        returning a tuple of an attribute name and the attribute for each element in the object.

        Returns:
            zip: A generator to be iterated over.
    """
    @requires_valid_data(default=iter(()))
    def items(self):
        for attr_name in self._all_attrs:
            if getattr(self, attr_name, None) != None:
                yield (attr_name, getattr(self,attr_name))

    """
    Similar to dict.keys, this method gives a list of the names of all the
    attributes in the object.

    Returns:
        list: The names of all the attributes in the object.
    """
    @requires_valid_data(default=[])
    def keys(self):
        return list(x for x in self)

    """
        Similar to dict.values, this method gives a list of the values of
        all the attributes in the object.

        Returns:
            list: The values of all the attributes in the object.
    """
    @requires_valid_data(default=[])
    def values(self):
        return list(y for x,y in self.items())


    """
        Gets the validity of the object, as outlined in each object's constructor.

        Returns:
            bool: True if the object is valid, otherwise False.
    """
    def is_valid(self):
        return self._is_valid

    """
        Converts the object to a dict with optionally being able to recurse of to_dict
        methods for composite AmzBase objects and to place these composite elements at
        the same level when flatten=True.

        Calling dict(cls) will have the same effect as calling
        cls.to_dict(recursive=False, flatten=False).

        Optional Args:
            recursive (bool): If true any values with a to_dict method will have their
              method called too, otherwise the value will be the object.
            flatten (bool): If true recursive calls to to_dict will cause the values to
              be merged at the top level. Note that recursive must be True for this to have an effect.

         Returns:
             dict: A dict of the with attribute names as keys and their values as values.
    """
    def to_dict(self, recursive=True, flatten=False):
        d = {}
        for k, v in self.items():
            if recursive == True and hasattr(v,'to_dict'):
                if flatten == True:
                    d = {**d, **v.to_dict()}
                else:
                    d[k] = v.to_dict()
            else:
                d[k] = v
        return d

    """
        Pandas must be installed for this method to be called. It will convert
        the object to a Pandas Series using the same possible recursive and
        flattening options of the to_dict method above.

        Optional Args:
            recursive (bool): See above.
            flatten (bool): See above.

        Returns:
            Pandas DataFrame: A series with attribute names as keys and their values as values.
    """
    def to_series(self, recursive=True, flatten=False):
        #only import at this point as amzSear can be used without pandas if desired
        from pandas import Series 
        return Series(self.to_dict(recursive=recursive,flatten=flatten))



import re

try:
    from amzsear.core.AmzBase import AmzBase
    from amzsear.core import requires_valid_data, capture_exception
except ImportError:
    from .amzsear.core.AmzBase import AmzBase
    from .amzsear.core import requires_valid_data, capture_exception


"""
    AmzRating(html_element=None):

    The AmzRating class extends the AmzBase class and, as such the following
    attributes are available to be called as an index call or as an attribute:

        ratings_text (str): The star rating (e.g. "4.5/5").
        ratings_count_text (str): The number of votes (e.g. "100").

    This class should usually not be instantiated directly (rather be used as
    part of an AmzProduct element) but can be created by passing an HTML element
    to the constructor. If nothing is passed, an empty AmzRating object is created.

    Optional Args:
        html_element (LXML root): A root for an HTML tree derived from an element
          on an Amazon search page.
"""
class AmzRating(AmzBase):
    ratings_text = None
    ratings_count_text = None
    _all_attrs = ['ratings_text', 'ratings_count_text']

    """
        Constructor takes an lxml Element and extracts (if possible)
          the first ratings text and ratings count text
    """
    def __init__(self, html_element=None):
        if html_element != None:
            (ratings_text, ratings_count_text) = self._get_from_html(html_element)

            # Values are only set if there are 2 ratings_text floats (the text should
            #  be of the form "N out of N stars") and one ratings_count float (the text
            #  should be of the form "N").

            if (len(self._extract_all_values(ratings_text)) == 2 and 
              len(self._extract_all_values(ratings_count_text)) == 1):

                self.ratings_text = ratings_text
                self.ratings_count_text = ratings_count_text
                self._is_valid = True


    """
        Private method - extracts the ratings text and count from the lxml element passed.

        Returns:
            tuple: Tuple of the ratings text followed by (a string of) the count
              for the rating. Both of these are strings.
    """
    @capture_exception(IndexError, ('',''))
    def _get_from_html(self, root):
        ratings_text = root.cssselect('i[class*="star"]')[0].text_content()
        ratings_count_text = root.cssselect('a[href*="customerReviews"]')[0].text_content()

        return (ratings_text,ratings_count_text)

    """
        Private method - extracts any numbers from some text.

        e.g. _extract_all_values("4.5 out of 5 stars") == [4.5, 5.0]

        Optional Args:
            data(str): The string to extract from. If not data is passed
              the ratings_text attribute is used.

        Returns:
            list: List of floats for any numbers in the text.

    """
    def _extract_all_values(self, data=None):
        if data == None:
            data = self.ratings_text

        return [float(re.sub('[^\d.]','',x)) for x in re.findall('[\d.,-]+',data)]

    """
        Gets a percentage value of the rating - 0% being a 0/5 star rating and
        100% being a 5/5 star rating.

        Returns:
            float: The star percentage.
    """
    @requires_valid_data(default=0)
    def get_perc(self):
        return self.get_numerator()/self.get_denominator()
 
    """
        Gets the value the average value of the star rating (usually between 0 and 5).

        Returns:
            float: The numerator of the star rating.
    """
    @requires_valid_data(default=0)
    def get_numerator(self):
        return sorted(self._extract_all_values())[0]

    """
        Gets the value the star rating is out of (usually 5).

        Returns:
            float: The denominator of the star rating.
    """ 
    @requires_valid_data(default=0)
    def get_denominator(self):
        return sorted(self._extract_all_values())[-1]

    """
        Gets the total number of ratings.

        Returns:
            int: The number of ratings.
    """ 
    @requires_valid_data(default=0)
    def get_count(self):
        return int(self._extract_all_values(self.ratings_count_text)[0])

    """
        Gives a visual representation of the star rating, rounding to the nearest star.

        Optional Args:
            star_repr(str): A (typically) single character used to represent a star.

        Returns:
            str: A representation of the star rating.
    """
    @requires_valid_data(default='')
    def get_star_repr(self,star_repr='*'):
        return star_repr*round(self.get_numerator())



from urllib import parse

try:
    from amzsear.core.consts import (QUERY_BUILD_DICT, BASE_URL, DEFAULT_REGION,
        REGION_CODES, SEARCH_URL)
except ImportError:
    from .amzsear.core.consts import (QUERY_BUILD_DICT, BASE_URL, DEFAULT_REGION,
        REGION_CODES, SEARCH_URL)


"""
    Decorator for valid data in an object, returns default if not valid
"""
def requires_valid_data(default=None):
    def funcParams(f):
        def funcWrapper(self,*args,**kws):
            if hasattr(self, '_is_valid') and self._is_valid == True:
                return f(self,*args,**kws)
            else:
                return default
        return funcWrapper
    return funcParams



"""
    Decorator to capture exception and pass a default instead
"""
def capture_exception(error, default=None):
    def funcParams(f):
        def funcWrapper(*args,**kws):
            try:
                return f(*args, **kws)
            except error:
                return default
        return funcWrapper
    return funcParams


"""
    Builds a url based on a query
"""
def build_url(url=None, query='', page_num=1, region=DEFAULT_REGION):
    if url == None:
        #build from query, page_num and region
        base = build_base_url(region)
        url = SEARCH_URL % (base, query, page_num)

    if url.startswith('/'):
        url =  build_base_url(region) + url 

    parsed_obj = parse.urlparse(url)
    query_dict = parse.parse_qs(parsed_obj.query)

    #update the query dict
    query_dict.update(QUERY_BUILD_DICT)

    parsed_obj = parsed_obj._replace(query=parse.urlencode(query_dict, doseq=True)) 
    return parsed_obj.geturl()


"""
    Builds URL based on region
"""
def build_base_url(region=DEFAULT_REGION):
    find_region = region.upper()
    if find_region not in REGION_CODES.keys():
        raise ValueError('%s is not a know Amazon region' % (repr(region)))

    return BASE_URL + REGION_CODES[find_region]



from setuptools import setup, find_packages
setup(
  name = 'amzsear',
  packages = find_packages(),
  version = '2.0.1',
  description = 'The unofficial Amazon search CLI & Python API',
  author = "Asher Silvers",
  author_email = "ashersilvers@gmail.com",
  license='MIT',
  url = 'https://github.com/asherAgs/amzSear', 
  keywords = 'amazon search product products python',
  classifiers = [],
  entry_points={
    'console_scripts': [
      'amzsear = amzsear.cli.cli:run',
     ],
  },
)



try:
    from amzsear import AmzSear
    from amzsear.core.consts import DEFAULT_REGION, REGION_CODES
except ImportError:
    from .amzsear import AmzSear
    from .amzsear.core.consts import DEFAULT_REGION, REGION_CODES

import argparse
import webbrowser
import json
import sys
import csv

"""
"""
def run(*passed_args):
    parser = get_parser()
    args = parser.parse_args(*passed_args) # the parser defaults to sys args if nothing passed
    args = vars(args)

    amz_args = {x:y for x,y in args.items() if x not in ['item','output','dont_open']}
    out = AmzSear(**amz_args)

    if args['item'] != None:
        # single item selection
        prod = out[args['item']]
        out = AmzSear(products=[prod]) # error raised if not found
        out._urls = [prod.product_url] 


    # handle output types
    if args['output'] == 'short':
        print_short(out)
    if args['output'] == 'verbose':
        print_verbose(out)
    elif args['output'] == 'csv':
        print_csv(out)
    elif args['output'] == 'json':
        print_json(out)
    # elif args['output'] == 'quite' --> no output


    if args['dont_open'] != True:
        for url in out._urls:
            webbrowser.open(url)



def get_parser():
    parser = argparse.ArgumentParser(description='The unofficial Amazon search CLI')

    parser.add_argument('query', type=str, help='The query string to be searched')
    parser.add_argument('-p','--page', type=int,
        help='The page number to be searched (defaults to 1)', default=1)
    parser.add_argument('-i','--item', type=str,
        help='The item index to be displayed (relative to the page)', default=None)
    parser.add_argument('-r','--region', type=str, choices=REGION_CODES,
        default=DEFAULT_REGION, help='The amazon country/region to be searched')

    parser.add_argument('-d','--dont-open', action='store_true',
        help='Stop the page from opening in the default browser')

    parser.add_argument('-o','--output', type=str, choices=['short','verbose','quiet','csv','json'],
        default='short', help='The output type to be displayed (defaults to short)')

    return parser


def print_csv(cls):
    # flattens to list of dicts with index value
    data = [{**v.to_dict(flatten=True),**({'_index' : k})} for k,v in cls.items()]

    # print with all quotes
    writer = csv.DictWriter(sys.stdout, data[0].keys(), quoting=csv.QUOTE_ALL) 
    writer.writeheader()
    writer.writerows(data)

def print_json(cls):
    print(json.dumps({k: v.to_dict() for k,v in cls.items()}))

def print_verbose(cls):
    print(cls)


def print_short(cls):
    fields = ['','Title','Prices','Rating']

    rows = [{f:f for f in fields}]
    for index, product in cls.items():
        temp_dict = {}

        temp_dict[''] = index

        # get price in format '$nn.nn-$mm.mm'
        price_tup = {product.prices[k]:product.get_prices(k) for k in product.prices}
        if len(price_tup) > 0:
            price_tup = (min(price_tup, key=lambda x: price_tup[x]), max(price_tup, key=lambda x: price_tup[x]))
            if price_tup[0] == price_tup[-1]:
                temp_dict['Prices'] = price_tup[0] # one price
            else:
                temp_dict['Prices'] = price_tup[0] + ' - ' + price_tup[-1] # range of prices
        else:
            temp_dict['Prices'] = '------------'
        temp_dict['Title'] = product.get('title','----------')[:50] # limit title length
    
        temp_dict['Rating'] = product.get('rating','-----')
        if temp_dict['Rating'] != '-----':
            temp_dict['Rating'] = temp_dict['Rating'].get_star_repr()

        rows.append(temp_dict)

    format_str = []
    for field in fields:
        #get longest in each field into format_str
        format_str.append('{:%d}' % (max(len(x[field]) for x in rows)+ 1))
    format_str = ' '.join(format_str)

    for row in rows:
        print(format_str.format(*[row.get(x,'') for x in fields])) # print in order

if __name__ == '__main__':
    run()



import re

try:
    from amzsear.core.AmzBase import AmzBase
    from amzsear.core import requires_valid_data, capture_exception, build_url
    from amzsear.core.AmzRating import AmzRating
except ImportError:
    from .amzsear.core.AmzBase import AmzBase
    from .amzsear.core import requires_valid_data, capture_exception, build_url
    from .amzsear.core.AmzRating import AmzRating
"""
    The AmzProduct class extends the AmzBase class and, as such the following
    attributes are available to be called as an index call or as an attribute:

        title (str): The name of the product.
        product_url (str) A url directly to the product's Amazon page.
        image_url (str) A url to the product's default image.
        rating (AmzRating) An AmzRating object.
        prices (dict) A dictionary of prices, with the price type as a key and
          a string for the price value (see get_prices method to get float values).
        extra_attributes (dict) Any extra information that can be extracted
          from the product.
        subtext (list) A list of strings under the title, typically the author's
          name and/or the date of publication.

    This class should usually not be instantiated directly (rather be used in
    an (AmzSear object) but can be created by passing an HTML element to
    the constructor. If nothing is passed, an empty AmzProduct object is
    created.

    Optional Args:
        html_element (LXML root): A root for an HTML tree derived from an element on an Amazon search page.
"""
class AmzProduct(AmzBase):
    title = None
    product_url = None
    image_url = None
    rating = None
    prices = None
    extra_attributes = None 
    subtext = None

    _all_attrs = ['title','product_url','image_url','rating','prices',
        'extra_attributes', 'subtext']

    def __init__(self, html_element=None):
        if html_element != None:
            html_dict = self._get_from_html(html_element)
            for k,v in html_dict.items():
                setattr(self, k, v)
            if len(html_dict) > 0:
                self._is_valid = True

    """
        Private method - used in to initialise fields from HTML

        Returns:
            dict: A dict of fields with extracted data
    """
    @capture_exception(IndexError,default={})
    def _get_from_html(self, root):
        d = {}

        title_root = [x for x in root.cssselect('a') if len(x.cssselect('h2')) > 0][0]
        d['title'] = ''.join([x.text_content() for x in title_root.cssselect('h2')])
        d['product_url'] = build_url(title_root.get('href'))
        for elem in title_root.getparent().getparent().cssselect('div[class="a-row a-spacing-none"]'):
            temp_subtext = ''.join([x.text_content() for x in elem.cssselect('span[class*="a-size-small"]')])
            if len(temp_subtext) > 0:
                d['subtext'] = d.get('subtext',[]) + [temp_subtext]


        d['image_url'] = root.cssselect('img[src]')[0].get('src')
        d['rating'] = AmzRating(root) or None

        d['prices'] = {}
        price_names = root.cssselect('h3[data-attribute]')
        price_text = root.cssselect('span[class^="a"]')
        price_text = filter(lambda x: re.match('^[^a-z\-]+$',str(x.text)) and
            re.search('[\.\,]',str(x.text)) and re.search('\d',str(x.text)), price_text)

        for i, el in enumerate(price_text):
            if i >= len(price_names):
                price_key = str(len(d['prices'])) # defaults to a number if no name for price type
            else:
                price_key = price_names[i].text
            d['prices'][price_key] = el.text

        extras = root.cssselect('div[class="a-fixed-left-grid-inner"] > div > span')
        extras = [re.sub('\s+',' ', x.text_content().strip()) for x in extras]
        d['extra_attributes'] = dict(list(zip(extras,extras[1:]))[::2])

        # _index is not used explicitly in _all_attrs but can be referenced elsewhere
        d['_index'] = root.get('id','').split('_')[-1]

        # clean up before returning
        return dict(map(lambda k: (k, d[k].strip() if isinstance(d[k],str) else d[k]), d)) 


    """
        Gets a list of floats from the dictionary of price text. A key can be passed
        to explicitly specify the prices to select. If the key is None, all prices keys
        are used.

        Optional Args:
            key (str or list): A key or list of keys to in the price dictionary.

        Returns:
            list: List of floats for the subset of price specified.
    """
    @requires_valid_data(default=[])
    def get_prices(self, key=None):
        keys = []
        if key == None:
            keys = self.prices.keys()
        elif isinstance(key,list):
            keys = key
        else:
            keys = [key]

        prices = []
        for k in keys:
            if k not in self.prices:
                raise KeyError(k)
            prices += [re.sub(',','',x) for x in re.findall('[\d.,]+', self.prices[k])]

        return sorted(map(float,prices))

